﻿'Programed by NAS6

Imports System.Windows


Public Class Local
    Implements ICloneable

    Public lcl As New MyQueue(Of Matrix)    '行列群
    Public bcmr As Boolean = False          'カメラ原点フラグ＠逆行列で実装すると浮動小数点丸め誤差が出るため


    Public Function Clone() As Object Implements ICloneable.Clone
        Dim ret As New Local
        ret.lcl = Me.lcl.Clone()
        ret.bcmr = Me.bcmr
        Return ret
    End Function

    Public Function Mat() As Matrix
        Dim IntWK As Integer
        Dim MatWK As New Matrix(4)
        MatWK = MatWK.UnitMat()
        For IntWK = 0 To lcl.Count - 1
            MatWK *= lcl.buf(IntWK)
        Next
        Return MatWK
    End Function
End Class

Public Class Material
    Implements ICloneable

    Public ka As Double                 '反射係数環境光
    Public kb As Double                 '反射係数その他
    Public bl As Boolean                '照明反転

    Public Sub New()
        ka = 1.0
        kb = 1.0
        bl = False
    End Sub

    Public Sub New(ByVal pka As Double, ByVal pkb As Double)
        ka = pka
        kb = pkb
        bl = False
    End Sub

    Public Sub New(ByVal pka As Double, ByVal pkb As Double, ByVal pbl As Boolean)
        ka = pka
        kb = pkb
        bl = pbl
    End Sub

    Public Function Clone() As Object Implements ICloneable.Clone
        Dim ret As New Material
        ret.ka = Me.ka
        ret.kb = Me.kb
        ret.bl = Me.bl
        Return ret
    End Function
End Class

Public Class Label
    Implements ICloneable

    Public offset As Point  '文字列左上座標
    Public str As Paragraph '文字列
    Public col As Color     '文字色

    Public Sub New()
        offset = New Point(0, 0)
        str = New Paragraph("")
        col = Color.White
    End Sub

    Public Sub New(ByVal pstr As String, ByVal pofs As Point, ByVal pcol As Color)
        offset = pofs
        str = New Paragraph(pstr)
        col = pcol
    End Sub
    Public Sub New(ByVal pstr As String, ByVal pofs As Point)
        offset = pofs
        str = New Paragraph(pstr)
        col = Color.White
    End Sub

    Public Sub New(ByVal pstr As String, ByVal pcol As Color)
        offset = New Point(0, 0)
        str = New Paragraph(pstr)
        col = pcol
    End Sub

    Public Sub New(ByVal pstr As String)
        offset = New Point(0, 0)
        str = New Paragraph(pstr)
        col = Color.White
    End Sub

    Public Function Clone() As Object Implements ICloneable.Clone
        Dim ret As New Label
        ret.offset = Me.offset
        ret.str = Me.str.Clone()
        ret.col = Me.col
        Return ret
    End Function
End Class

Public Class Triangle
    Implements ICloneable

    Public v(2) As Vector               '頂点
    Public n As Vector                  '法線
    Public vcol As Color                '頂点カラー
    Public tcol As Color                '面カラー
    Public mtr As Material              '物性

    Public Sub New()
        Dim IntWK As Integer
        For IntWK = 0 To 2
            v(IntWK) = New Vector(4, True)
        Next
        n = New Vector(4, True)
        vcol = Color.White
        tcol = Color.Black
        mtr = New Material
    End Sub

    Public Sub New(ByVal pv() As Vector, ByVal pn As Vector, ByVal pvc As Color, ByVal ptc As Color, ByVal pm As Material)
        Dim IntWK As Integer
        For IntWK = 0 To 2
            v(IntWK) = pv(IntWK).Clone()
        Next
        n = pn.Clone()
        vcol = pvc
        tcol = ptc
        mtr = pm.Clone()
    End Sub

    Public Function Clone() As Object Implements ICloneable.Clone
        Dim ret As New Triangle
        Dim i As Integer
        For i = 0 To 2
            ret.v(i) = Me.v(i).Clone()
        Next
        ret.n = Me.n.Clone()
        ret.vcol = Me.vcol
        ret.tcol = Me.tcol
        ret.mtr = Me.mtr.Clone()
        Return ret
    End Function
End Class

Public Class Polygon
    Implements IComparable
    Implements ICloneable

    Public Local As New Local       'ローカル行列群
    Public v() As Triangle          '三角形リスト
    Public Range As Double = 1.0    '範囲
    Public Base As Double = 2.0     '基数
    Public Label As New Label       'ラベル


    Public Function Clone() As Object Implements ICloneable.Clone
        Dim i As Integer
        Dim vv(v.Length - 1) As Triangle
        For i = 0 To v.Length - 1
            vv(i) = v(i).Clone()
        Next
        Dim ret As New Polygon(Me.Local.Clone(), vv, Me.Range, Me.Base, Me.Label.Clone())
        Return ret
    End Function

    Public Function CompareTo(obj As Object) As Integer Implements IComparable.CompareTo
        CompareTo = 0
        Dim lh As Polygon = Me
        Dim rh As Polygon = obj
        If Not (lh.Base = rh.Base) Then
            CompareTo = CompareTo Or 1
        End If
    End Function

    Public Overrides Function Equals(obj As Object) As Boolean
        Dim ret As Integer = CompareTo(obj)
        If ret = 0 Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Shared Operator =(ByVal lh As Polygon, ByVal rh As Polygon) As Boolean
        Dim ret As Integer = lh.CompareTo(rh)
        If ret = 0 Then
            Return True
        Else
            Return False
        End If
    End Operator

    Public Shared Operator <>(ByVal lh As Polygon, ByVal rh As Polygon) As Boolean
        Return (Not lh = rh)
    End Operator


    'p():TRIANGLELIST
    Public Sub New(ByVal lcl As Local, ByVal p() As Triangle)
        Local = lcl.Clone()
        Dim i As Integer
        Dim nn As Integer = p.Length - 1
        ReDim v(nn)
        For i = 0 To nn
            v(i) = p(i).Clone()
        Next
        Dim vec1 As Vector
        Dim vec2 As Vector
        Dim vec3 As Vector
        Dim vec21 As Vector
        Dim vec31 As Vector
        For i = 0 To nn
            vec1 = v(i).v(0).Clone()
            vec2 = v(i).v(1).Clone()
            vec3 = v(i).v(2).Clone()
            vec21 = vec2 - vec1
            vec31 = vec3 - vec1
            v(i).n = vec21.Cross(vec31).NormalVec()
            v(i).mtr.bl = Not v(i).mtr.bl
        Next
    End Sub

    'p():TRIANGLELIST
    Public Sub New(ByVal lcl As Local, ByVal p() As Triangle, ByVal rg As Double, ByVal bs As Double)
        Local = lcl.Clone()
        Range = rg
        Base = bs
        Dim i As Integer
        Dim nn As Integer = p.Length - 1
        ReDim v(nn)
        For i = 0 To nn
            v(i) = p(i).Clone()
        Next
        Dim vec1 As Vector
        Dim vec2 As Vector
        Dim vec3 As Vector
        Dim vec21 As Vector
        Dim vec31 As Vector
        For i = 0 To nn
            vec1 = v(i).v(0).Clone()
            vec2 = v(i).v(1).Clone()
            vec3 = v(i).v(2).Clone()
            vec21 = vec2 - vec1
            vec31 = vec3 - vec1
            v(i).n = vec21.Cross(vec31).NormalVec()
        Next
    End Sub


    'p():TRIANGLELIST
    Public Sub New(ByVal lcl As Local, ByVal p() As Triangle, ByVal lbl As Label)
        Local = lcl.Clone()
        Label = lbl.Clone()
        Dim i As Integer
        Dim nn As Integer = p.Length - 1
        ReDim v(nn)
        For i = 0 To nn
            v(i) = p(i).Clone()
        Next
        Dim vec1 As Vector
        Dim vec2 As Vector
        Dim vec3 As Vector
        Dim vec21 As Vector
        Dim vec31 As Vector
        For i = 0 To nn
            vec1 = v(i).v(0).Clone()
            vec2 = v(i).v(1).Clone()
            vec3 = v(i).v(2).Clone()
            vec21 = vec2 - vec1
            vec31 = vec3 - vec1
            v(i).n = vec21.Cross(vec31).NormalVec()
            v(i).mtr.bl = Not v(i).mtr.bl
        Next
    End Sub

    'p():TRIANGLELIST
    Public Sub New(ByVal lcl As Local, ByVal p() As Triangle, ByVal rg As Double, ByVal bs As Double, ByVal lbl As Label)
        Local = lcl.Clone()
        Label = lbl.Clone()
        Range = rg
        Base = bs
        Dim i As Integer
        Dim nn As Integer = p.Length - 1
        ReDim v(nn)
        For i = 0 To nn
            v(i) = p(i).Clone()
        Next
        Dim vec1 As Vector
        Dim vec2 As Vector
        Dim vec3 As Vector
        Dim vec21 As Vector
        Dim vec31 As Vector
        For i = 0 To nn
            vec1 = v(i).v(0).Clone()
            vec2 = v(i).v(1).Clone()
            vec3 = v(i).v(2).Clone()
            vec21 = vec2 - vec1
            vec31 = vec3 - vec1
            v(i).n = vec21.Cross(vec31).NormalVec()
        Next
    End Sub
End Class
