﻿'Programed by NAS6


'クォータニオン
Public Class Quaternion
    Implements IComparable
    Implements ICloneable

    Public q As New Vector(4)


    Public Function Clone() As Object Implements ICloneable.Clone
        Dim ret As New Quaternion
        ret.q = Me.q.Clone()
        Return ret
    End Function

    Public Function CompareTo(obj As Object) As Integer Implements IComparable.CompareTo
        CompareTo = 0
        Dim lh As Quaternion = Me
        Dim rh As Quaternion = obj
        Dim i As Integer
        For i = 0 To 3
            If Not lh.q.x(i) = rh.q.x(i) Then
                CompareTo = CompareTo Or (1 << i)
            End If
        Next
    End Function

    Public Overrides Function Equals(obj As Object) As Boolean
        Dim ret As Integer = CompareTo(obj)
        If ret = 0 Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Shared Operator =(ByVal lh As Quaternion, ByVal rh As Quaternion) As Boolean
        Dim ret As Integer = lh.CompareTo(rh)
        If ret = 0 Then
            Return True
        Else
            Return False
        End If
    End Operator

    Public Shared Operator <>(ByVal lh As Quaternion, ByVal rh As Quaternion) As Boolean
        Return (Not lh = rh)
    End Operator


    Public Sub New()
        Dim i As Integer
        For i = 0 To 3
            q.x(i) = 0.0
        Next
    End Sub

    Public Sub New(ByVal w As Double, ByVal x As Double, ByVal y As Double, ByVal z As Double)
        q.x(0) = w
        q.x(1) = x
        q.x(2) = y
        q.x(3) = z
    End Sub

    Public Sub New(ByVal s As Double, ByVal v As Vector)
        q.x(0) = s
        q.x(1) = v.x(0)
        q.x(2) = v.x(1)
        q.x(3) = v.x(2)
    End Sub

    Public Shared Operator *(ByVal lh As Quaternion, ByVal rh As Double) As Quaternion
        Dim IntWK As Integer = 0
        Dim QuatWK As New Quaternion
        For IntWK = 0 To 3
            QuatWK.q.x(IntWK) = lh.q.x(IntWK) * rh
        Next
        Return QuatWK
    End Operator

    Public Shared Operator /(ByVal lh As Quaternion, ByVal rh As Double) As Quaternion
        Dim IntWK As Integer = 0
        Dim QuatWK As New Quaternion
        If rh = 0.0 Then
            Return QuatWK
        End If
        For IntWK = 0 To 3
            QuatWK.q.x(IntWK) = lh.q.x(IntWK) / rh
        Next
        Return QuatWK
    End Operator

    Public Shared Operator +(ByVal lh As Quaternion, ByVal rh As Quaternion) As Quaternion
        Dim IntWK As Integer = 0
        Dim QuatWK As New Quaternion
        For IntWK = 0 To 3
            QuatWK.q.x(IntWK) = lh.q.x(IntWK) + rh.q.x(IntWK)
        Next
        Return QuatWK
    End Operator

    Public Shared Operator -(ByVal lh As Quaternion, ByVal rh As Quaternion) As Quaternion
        Dim IntWK As Integer = 0
        Dim QuatWK As New Quaternion
        For IntWK = 0 To 3
            QuatWK.q.x(IntWK) = lh.q.x(IntWK) - rh.q.x(IntWK)
        Next
        Return QuatWK
    End Operator

    Public Shared Operator *(ByVal lh As Quaternion, ByVal rh As Quaternion) As Quaternion
        Dim QuatWK As New Quaternion
        QuatWK.q.x(0) = lh.q.x(0) * rh.q.x(0) - lh.q.x(1) * rh.q.x(1) - lh.q.x(2) * rh.q.x(2) - lh.q.x(3) * rh.q.x(3)
        QuatWK.q.x(1) = lh.q.x(0) * rh.q.x(1) + lh.q.x(1) * rh.q.x(0) + lh.q.x(2) * rh.q.x(3) - lh.q.x(3) * rh.q.x(2)
        QuatWK.q.x(2) = lh.q.x(0) * rh.q.x(2) - lh.q.x(1) * rh.q.x(3) + lh.q.x(2) * rh.q.x(0) + lh.q.x(3) * rh.q.x(1)
        QuatWK.q.x(3) = lh.q.x(0) * rh.q.x(3) + lh.q.x(1) * rh.q.x(2) - lh.q.x(2) * rh.q.x(1) + lh.q.x(3) * rh.q.x(0)
        Return QuatWK
    End Operator

    '絶対値2乗
    Public Function SquareAbs() As Double
        Dim IntWK As Integer = 0
        Dim DblWK As Double = 0.0
        Dim QuatWK As New Quaternion
        QuatWK = Me.Clone()
        For IntWK = 0 To 3
            DblWK += QuatWK.q.x(IntWK) * QuatWK.q.x(IntWK)
        Next
        Return DblWK
    End Function

    '絶対値
    Public Function Abs() As Double
        Dim DblWK As Double = 0.0
        DblWK = Math.Sqrt(Me.SquareAbs())
        Return DblWK
    End Function

    '共役
    Public Function Conjugation() As Quaternion
        Dim QuatWK As New Quaternion(Me.q.x(0), -Me.q.x(1), -Me.q.x(2), -Me.q.x(3))
        Return QuatWK
    End Function

    '逆元
    Public Function Inverse() As Quaternion
        Dim QuatWK As New Quaternion
        QuatWK = Me.Conjugation() / Me.SquareAbs()
        Return QuatWK
    End Function

    'ノーマライズ
    Public Function Normal() As Quaternion
        Dim QuatWK As New Quaternion
        QuatWK = Me / Me.Abs()
        Return QuatWK
    End Function

    '内積
    Public Function Dot(ByVal rh As Quaternion) As Double
        Dim sum As Double = 0.0
        Dim i As Integer
        For i = 0 To 3
            sum = sum + Me.q.x(i) * rh.q.x(i)
        Next
        Dot = sum
    End Function

    '軸に対する回転
    Public Function RotAxisQuat(ByVal Axis As Vector, ByVal theta As Double) As Quaternion
        Dim IntWK As Integer = 0
        Dim QuatWK As New Quaternion
        If Not Axis.x.Length = 4 Then
            Return QuatWK
        End If
        Dim VecWK As Vector = Axis.Homogeneous().NormalVec()
        QuatWK.q.x(IntWK) = Math.Cos(theta / 2.0)
        For IntWK = 1 To 3
            QuatWK.q.x(IntWK) = VecWK.x(IntWK) * Math.Sin(theta / 2.0)
        Next
        Return QuatWK
    End Function

    '回転
    Public Function Matrix() As Matrix
        Dim MatWK As New Matrix(4)
        Dim q As New Quaternion
        q = Me.Clone()

        MatWK.x(0).x(0) = 1.0
        MatWK.x(0).x(1) = 0.0
        MatWK.x(0).x(2) = 0.0
        MatWK.x(0).x(3) = 0.0

        MatWK.x(1).x(0) = 0.0
        MatWK.x(1).x(1) = 1.0 - 2.0 * q.q.x(2) * q.q.x(2) - 2.0 * q.q.x(3) * q.q.x(3)
        MatWK.x(1).x(2) = 2.0 * q.q.x(1) * q.q.x(2) - 2.0 * q.q.x(0) * q.q.x(3)
        MatWK.x(1).x(3) = 2.0 * q.q.x(1) * q.q.x(3) + 2.0 * q.q.x(0) * q.q.x(2)

        MatWK.x(2).x(0) = 0.0
        MatWK.x(2).x(1) = 2.0 * q.q.x(2) * q.q.x(1) + 2.0 * q.q.x(0) * q.q.x(3)
        MatWK.x(2).x(2) = 1.0 - 2.0 * q.q.x(1) * q.q.x(1) - 2.0 * q.q.x(3) * q.q.x(3)
        MatWK.x(2).x(3) = 2.0 * q.q.x(2) * q.q.x(3) - 2.0 * q.q.x(0) * q.q.x(1)

        MatWK.x(3).x(0) = 0.0
        MatWK.x(3).x(1) = 2.0 * q.q.x(3) * q.q.x(1) - 2.0 * q.q.x(0) * q.q.x(2)
        MatWK.x(3).x(2) = 2.0 * q.q.x(3) * q.q.x(2) + 2.0 * q.q.x(0) * q.q.x(1)
        MatWK.x(3).x(3) = 1.0 - 2.0 * q.q.x(1) * q.q.x(1) - 2.0 * q.q.x(2) * q.q.x(2)

        Return MatWK
    End Function

    '線形補完
    Public Function Lerp(ByVal q As Quaternion, ByVal t As Double) As Quaternion
        Dim ret As New Quaternion
        Dim i As Integer
        For i = 0 To 3
            ret.q.x(i) = (1.0 - t) * Me.q.x(i) + t * q.q.x(i)
        Next
        Lerp = ret / ret.Abs()
    End Function

    '球面線形補完
    Public Function Slerp(ByVal q As Quaternion, ByVal t As Double) As Quaternion
        Dim QuatWK As New Quaternion
        Dim d As Double = Me.Dot(q)
        Dim s As Double = 1.0 - (d * d)
        Dim ph As Double = Math.Acos(d)
        Dim s1 As Double
        Dim s2 As Double
        Dim sgn As Double = 1.0
        If s = 0.0 Then
            QuatWK = Me.Clone()
        Else
            If d < 0.0 Then
                If Math.PI / 2.0 < ph Then
                    sgn = -1.0
                End If
                d *= -1.0
                ph = Math.Acos(d)
                s1 = Math.Sin(ph * (1.0 - t)) / Math.Sin(ph)
                s2 = Math.Sin(ph * t) / Math.Sin(ph)
            Else
                s1 = Math.Sin(ph * (1.0 - t)) / Math.Sin(ph)
                s2 = Math.Sin(ph * t) / Math.Sin(ph)
            End If
        End If
        QuatWK.q.x(0) = (Me.q.x(0) * s1 + q.q.x(0) * s2) * sgn
        QuatWK.q.x(1) = (Me.q.x(1) * s1 + q.q.x(1) * s2) * sgn
        QuatWK.q.x(2) = (Me.q.x(2) * s1 + q.q.x(2) * s2) * sgn
        QuatWK.q.x(3) = (Me.q.x(3) * s1 + q.q.x(3) * s2) * sgn
        Return QuatWK
    End Function

    '球面線形補完
    Public Function Slerp2(ByVal q As Quaternion, ByVal t As Double) As Quaternion
        Dim d As Double = Me.Dot(q)
        Dim s As Double = Math.Sqrt(1.0 - d * d)
        Dim QuatWK As New Quaternion
        If s = 0 Then
            Return QuatWK
        End If
        QuatWK = Me * (s * (1.0 - t) / s) + q * ((s * t) / s)
        Return QuatWK
    End Function

    '4次元球
    Public Function Sphere4D() As Vector
        Sphere4D = New Vector(4, True)
        Dim i As Integer
        Dim d As Double = Math.Sqrt(2.0 * (1.0 - Me.q.x(0)))
        Sphere4D.x(0) = (1.0 - Me.q.x(0)) / d
        For i = 1 To 3
            Sphere4D.x(i) = Me.q.x(i) / d
        Next
    End Function

End Class
