﻿'Programed by NAS6



Imports System.Globalization
Imports System.Windows



Public Enum Edge As Integer
    LEFT = 1
    RIGHT = 2
    TOP = 4
    BOTTOM = 8
End Enum

'照明
Public Class Light
    Implements ICloneable

    Public acol As Color = Color.White                  '環境光
    Public px As Local                                  '点光源位置
    Public pcol As Color = Color.White                  '点光源
    Public kd(2) As Double                              '減衰係数
    Public dx As New Vector(4, True)                    '平行光源方向
    Public dcol As Color = Color.White                  '平行光源
    Public enable As Boolean = False                    '照明ON/OFF

    Public Mat As new Matrix(4)


    Public Function Clone() As Object Implements ICloneable.Clone
        Dim ret As New Light
        ret.acol = Me.acol
        ret.px = Me.px.Clone()
        ret.pcol = Me.pcol
        ret.kd = Me.kd.Clone()
        ret.dx = Me.dx.Clone()
        ret.dcol = Me.dcol
        ret.enable = Me.enable
        ret.Mat = Me.Mat.Clone()
        Return ret
    End Function

    Public Sub New()
        dx = dx.ZeroVec()
        dx.x(0) = 1.0

        Mat = Mat.UnitMat()
    End Sub

    Public Sub New(ByVal pac As Color, ByVal ppc As Color, ByVal ppx As Local, ByVal k() As Double, ByVal pdc As Color, ByVal pdx As Vector, ByVal pb As Boolean)
        If k.Length < 3 Then
            Return
        End If
        acol = pac
        Dim i As Integer
        pcol = ppc
        px = ppx.Clone()
        For i = 0 To 2
            kd(i) = k(i)
        Next
        dcol = pdc
        dx = pdx.Clone()
        enable = pb

        Mat = px.Mat()
    End Sub

    '減衰係数取得
    Public Function GetAttenuation(ByVal a As Integer) As Double
        If 0 < a And a < 3 Then
            GetAttenuation = kd(a)
        Else
            GetAttenuation = 0.0
        End If
    End Function

    '描画色取得
    Public Function GetColor(ByVal tcol As Color, ByVal n As Vector, ByVal p As Vector, ByVal mtr As Material, ByVal m As Matrix, ByVal bbb As Boolean) As Color
        If Not enable Or n.x(0) = 0.0 Then
            Return tcol
        End If

        Dim v As Vector = Mat.Pos()


        Dim pxwk As Vector = m * v
        Dim dpx As Vector
        If bbb Then
            If Not mtr.bl Then
                dpx = p - pxwk
            Else
                dpx = pxwk - p
            End If
        Else
            If mtr.bl Then
                dpx = p - pxwk
            Else
                dpx = pxwk - p
            End If
        End If
        Dim th As Double
        th = dpx.Theta(n) / (Math.PI / 2.0)
        If Double.IsNaN(th) Then
            th = 1.0
        End If
        If dpx.Abs() = 0.0 Then
            th = 1.0
        End If

        Dim k0 As Double = GetAttenuation(0)
        Dim k1 As Double = GetAttenuation(1)
        Dim k2 As Double = GetAttenuation(2)

        Dim d As Double = dpx.Abs()
        Dim atten As Double = 1.0 / (k0 + k1 * d + k2 * d * d)
        If Double.IsNaN(atten) Then
            atten = 1.0
        End If
        If 1.0 < atten Then
            atten = 1.0
        End If

        '環境光
        Dim ca1 As Color = Color.Black
        Dim cp1 As Color = Color.Black
        Dim cd1 As Color = Color.Black
        GetColor = Color.Black
        Dim r As Double
        Dim g As Double
        Dim b As Double
        r = CType((CType(acol.R, Double) / 255.0) * (CType(tcol.R, Double) / 255.0) * mtr.ka * 255.0, Integer)
        If 255 < r Then
            r = 255
        End If
        g = CType((CType(acol.G, Double) / 255.0) * (CType(tcol.G, Double) / 255.0) * mtr.ka * 255.0, Integer)
        If 255 < g Then
            g = 255
        End If
        b = CType((CType(acol.B, Double) / 255.0) * (CType(tcol.B, Double) / 255.0) * mtr.ka * 255.0, Integer)
        If 255 < b Then
            b = 255
        End If
        ca1.A = 255
        ca1.R = r
        ca1.G = g
        ca1.B = b

        '平行光源
        r = CType((CType(pcol.R, Double) / 255.0) * (CType(tcol.R, Double) / 255.0) * mtr.kb * atten * th * 255.0, Integer)
        If 255 < r Then
            r = 255
        End If
        g = CType((CType(pcol.G, Double) / 255.0) * (CType(tcol.G, Double) / 255.0) * mtr.kb * atten * th * 255.0, Integer)
        If 255 < g Then
            g = 255
        End If
        b = CType((CType(pcol.B, Double) / 255.0) * (CType(tcol.B, Double) / 255.0) * mtr.kb * atten * th * 255.0, Integer)
        If 255 < b Then
            b = 255
        End If
        cp1.A = 255
        cp1.R = r
        cp1.G = g
        cp1.B = b

        '点光源
        th = dx.Theta(n) / (Math.PI / 2.0)
        If Double.IsNaN(th) Then
            th = 1.0
        End If
        If dx.Abs() = 0.0 Then
            th = 1.0
        End If
        r = CType((CType(dcol.R, Double) / 255.0) * (CType(tcol.R, Double) / 255.0) * mtr.kb * th * 255.0, Integer)
        If 255 < r Then
            r = 255
        End If
        g = CType((CType(dcol.G, Double) / 255.0) * (CType(tcol.G, Double) / 255.0) * mtr.kb * th * 255.0, Integer)
        If 255 < g Then
            g = 255
        End If
        b = CType((CType(dcol.B, Double) / 255.0) * (CType(tcol.B, Double) / 255.0) * mtr.kb * th * 255.0, Integer)
        If 255 < b Then
            b = 255
        End If
        cd1.A = 255
        cd1.R = r
        cd1.G = g
        cd1.B = b

        '光源の和
        r = CType(((CType(ca1.R, Double) / 255.0) + (CType(cp1.R, Double) / 255.0) + (CType(cd1.R, Double) / 255.0)) * 255.0, Integer)
        If 255 < r Then
            r = 255
        End If
        g = CType(((CType(ca1.G, Double) / 255.0) + (CType(cp1.G, Double) / 255.0) + (CType(cd1.G, Double) / 255.0)) * 255.0, Integer)
        If 255 < g Then
            g = 255
        End If
        b = CType(((CType(ca1.B, Double) / 255.0) + (CType(cp1.B, Double) / 255.0) + (CType(cd1.B, Double) / 255.0)) * 255.0, Integer)
        If 255 < b Then
            b = 255
        End If
        GetColor.A = 255
        GetColor.R = r
        GetColor.G = g
        GetColor.B = b
    End Function

End Class

Public Class Render3D
    Public Const LineDelta As Double = 0.1
    Public Line As New MyQueue(Of Line)
    Public Poly As New MyQueue(Of Polygon)
    Public Sphr As New MyQueue(Of Sphere)
    Public ZBuf(,) As Double
    Public Range As Double = 1.0        '範囲
    Public Base As Double = 2.0         '基数
    Public Camera As New Local          'カメラ行列群
    Public CameraMat As New Matrix(4)   'カメラ行列

    Public BInf As Boolean = True
    Public Frs As Frustum

    Public lgt As New Light()

    Public Sub New()
    End Sub



    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal fr As Frustum)
        Camera = c.Clone()
        Line = ln.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal fr As Frustum, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal fr As Frustum)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal fr As Frustum, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Sphere), ByVal fr As Frustum)
        Camera = c.Clone()
        Line = ln.Clone()
        Sphr = p.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Sphere), ByVal fr As Frustum, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Sphr = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal fr As Frustum)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal fr As Frustum, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal fr As Frustum)
        Camera = c.Clone()
        Poly = p.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal fr As Frustum, ByVal l As Light)
        Camera = c.Clone()
        Poly = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Sphere), ByVal fr As Frustum)
        Camera = c.Clone()
        Sphr = p.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Sphere), ByVal fr As Frustum, ByVal l As Light)
        Camera = c.Clone()
        Sphr = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal fr As Frustum)
        Camera = c.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal fr As Frustum, ByVal l As Light)
        Camera = c.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
        BInf = False
        Frs = fr.Clone()
    End Sub




    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line))
        Camera = c.Clone()
        Line = ln.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal rg As Double, ByVal bs As Double)
        Camera = c.Clone()
        Line = ln.Clone()
        Range = rg
        Base = bs
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal rg As Double, ByVal bs As Double, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Range = rg
        Base = bs
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon))
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal rg As Double, ByVal bs As Double)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Range = rg
        Base = bs
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal rg As Double, ByVal bs As Double, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Range = rg
        Base = bs
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub


    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Sphere))
        Camera = c.Clone()
        Line = ln.Clone()
        Sphr = p.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double)
        Camera = c.Clone()
        Line = ln.Clone()
        Sphr = p.Clone()
        Range = rg
        Base = bs
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Sphere), ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Sphr = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Sphr = p.Clone()
        Range = rg
        Base = bs
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub


    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere))
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        Range = rg
        Base = bs
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal ln As MyQueue(Of Line), ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double, ByVal l As Light)
        Camera = c.Clone()
        Line = ln.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        Range = rg
        Base = bs
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub


    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon))
        Camera = c.Clone()
        Poly = p.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal rg As Double, ByVal bs As Double)
        Camera = c.Clone()
        Poly = p.Clone()
        Range = rg
        Base = bs
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal l As Light)
        Camera = c.Clone()
        Poly = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal rg As Double, ByVal bs As Double, ByVal l As Light)
        Camera = c.Clone()
        Poly = p.Clone()
        Range = rg
        Base = bs
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub


    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Sphere))
        Camera = c.Clone()
        Sphr = p.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double)
        Camera = c.Clone()
        Sphr = p.Clone()
        Range = rg
        Base = bs
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Sphere), ByVal l As Light)
        Camera = c.Clone()
        Sphr = p.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double, ByVal l As Light)
        Camera = c.Clone()
        Sphr = p.Clone()
        Range = rg
        Base = bs
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub


    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere))
        Camera = c.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double)
        Camera = c.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        Range = rg
        Base = bs
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal l As Light)
        Camera = c.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    Public Sub New(ByVal c As Local, ByVal p As MyQueue(Of Polygon), ByVal pp As MyQueue(Of Sphere), ByVal rg As Double, ByVal bs As Double, ByVal l As Light)
        Camera = c.Clone()
        Poly = p.Clone()
        Sphr = pp.Clone()
        Range = rg
        Base = bs
        lgt = l.Clone()
        CameraMat = CameraMat.UnitMat()
    End Sub

    '注視
    Public Function LookAtMat(ByVal eye As Vector, ByVal lookat As Vector, ByVal up As Vector) As Matrix
        Dim trns As Matrix = CameraMat.TransposedMat()
        Dim ep As Vector = trns.x(0).Clone()
        ep.bHomo = True

        Dim ez As Vector = lookat - eye
        ez = ez.NormalVec()

        Dim ex As Vector = up.Cross(ez)
        ex = ex.NormalVec()

        Dim ey As Vector = ez.Cross(ex)
        ey = ey.NormalVec()

        Dim rmat As New Matrix(4)
        rmat = rmat.UnitMat()
        rmat.x(1).x(1) = ex.x(1)
        rmat.x(1).x(2) = ex.x(2)
        rmat.x(1).x(3) = ex.x(3)

        rmat.x(2).x(1) = ey.x(1)
        rmat.x(2).x(2) = ey.x(2)
        rmat.x(2).x(3) = ey.x(3)

        rmat.x(3).x(1) = ez.x(1)
        rmat.x(3).x(2) = ez.x(2)
        rmat.x(3).x(3) = ez.x(3)

        Dim dt As Double
        LookAtMat = rmat.InverseMat(dt)
        LookAtMat.x(1).x(0) = ep.x(1)
        LookAtMat.x(2).x(0) = ep.x(2)
        LookAtMat.x(3).x(0) = ep.x(3)
    End Function

    '注視
    Public Function LookAtMat(ByVal ni As Integer, ByVal nb As Boolean) As Matrix
        Dim m As New Matrix(4)
        Dim mm As New Matrix(4)
        If nb Then
            Dim s As Sphere = Sphr.buf(ni)
            mm = s.Local.Mat()
        Else
            Dim p As Polygon = Poly.buf(ni)
            mm = p.Local.Mat()
        End If


        m = Camera.Mat()
        m *= CameraMat


        Dim eye As Vector = m.Pos()
        eye = eye.Homogeneous()
        Dim lookat As Vector = mm.Pos()
        lookat = lookat.Homogeneous()

        Dim nez As Vector = lookat - eye
        nez = nez.NormalVec()
        Dim axis As Vector
        Dim ax As New Vector(4, True)
        ax = ax.UnitVec(1)
        Dim az As New Vector(4, True)
        az = az.UnitVec(3)
        If Not (nez = ax Or nez = ax * -1.0) Then
            axis = ax.Clone()
        Else
            axis = az.Clone()
        End If

        Dim up As Vector
        up = nez.Cross(axis)
        up = up.Homogeneous().NormalVec()

        Return LookAtMat(eye, lookat, up)
    End Function

    '注視
    Public Function LookAtMat(ByVal v As Vector) As Matrix
        Dim m As New Matrix(4)
        m = Camera.Mat()
        m *= CameraMat


        Dim eye As Vector = m.Pos()
        eye = eye.Homogeneous()
        Dim lookat As Vector = v.Clone()
        lookat = lookat.Homogeneous()

        Dim nez As Vector = lookat - eye
        nez = nez.NormalVec()
        Dim axis As Vector
        Dim ax As New Vector(4, True)
        ax = ax.UnitVec(1)
        Dim az As New Vector(4, True)
        az = az.UnitVec(3)
        If Not (nez = ax Or nez = ax * -1.0) Then
            axis = ax.Clone()
        Else
            axis = az.Clone()
        End If

        Dim up As Vector
        up = nez.Cross(axis)
        up = up.Homogeneous().NormalVec()

        Return LookAtMat(eye, lookat, up)
    End Function

    '注視
    Public Function LookAtMat2(ByVal v As Vector) As Matrix
        Dim m As New Matrix(4)
        Dim mm As New Matrix(4)
        m = Camera.Mat()
        m *= CameraMat

        Dim dt As Double
        m = m.InverseMat(dt)
        mm = mm.UnitMat()
        mm.x(1).x(0) = v.x(1)
        mm.x(2).x(0) = v.x(2)
        mm.x(3).x(0) = v.x(3)
        m = m * mm

        Dim eye As Vector = m.Pos() * -1.0
        eye = eye.Homogeneous()
        Dim lookat As Vector = v.Clone()
        lookat = lookat.Homogeneous()

        Dim nez As Vector = lookat - eye
        nez = nez.NormalVec()
        Dim axis As Vector
        Dim ax As New Vector(4, True)
        ax = ax.UnitVec(1)
        Dim az As New Vector(4, True)
        az = az.UnitVec(3)
        If Not (nez = ax Or nez = ax * -1.0) Then
            axis = ax.Clone()
        Else
            axis = az.Clone()
        End If

        Dim up As Vector
        up = nez.Cross(axis)
        up = up.Homogeneous().NormalVec()

        Return LookAtMat(eye, lookat, up)
    End Function

    '線分クリップ
    Public Function SeqCode(ByVal p As Vector, ByVal min As Vector, ByVal max As Vector) As Integer
        SeqCode = 0
        If p.x(0) < min.x(0) Then
            SeqCode += Edge.LEFT
        End If
        If max.x(0) < p.x(0) Then
            SeqCode += Edge.RIGHT
        End If
        If p.x(1) < min.x(1) Then
            SeqCode += Edge.TOP
        End If
        If max.x(1) < p.x(1) Then
            SeqCode += Edge.BOTTOM
        End If
    End Function

    '線分クリップ
    Public Function IntsecX(ByVal p0 As Vector, ByVal p1 As Vector, ByVal clipx As Double, ByVal min As Vector, ByVal max As Vector, ByRef p As Vector) As Boolean
        Dim cy As Double = (p1.x(1) - p0.x(1)) * (clipx - p0.x(0)) / (p1.x(0) - p0.x(0)) + p0.x(1)
        If Double.IsNaN(cy) Or Double.IsInfinity(cy) Then
            cy = (p1.x(1) - p0.x(1)) * (clipx - p0.x(0))
        End If
        If cy < min.x(1) Or max.x(1) < cy Then
            Return False
        End If
        p.x(0) = clipx
        p.x(1) = cy
        Return True
    End Function

    '線分クリップ
    Public Function IntsecY(ByVal p0 As Vector, ByVal p1 As Vector, ByVal clipy As Double, ByVal min As Vector, ByVal max As Vector, ByRef p As Vector) As Boolean
        Dim cx As Double = (p1.x(0) - p0.x(0)) * (clipy - p0.x(1)) / (p1.x(1) - p0.x(1)) + p0.x(0)
        If Double.IsNaN(cx) Or Double.IsInfinity(cx) Then
            cx = (p1.x(0) - p0.x(0)) * (clipy - p0.x(1))
        End If
        If cx < min.x(0) Or max.x(0) < cx Then
            Return False
        End If
        p.x(0) = cx
        p.x(1) = clipy
        Return True
    End Function

    '線分クリップ
    Public Function LineClippedPoint(ByVal code As Edge, ByVal p0 As Vector, p1 As Vector, ByVal min As Vector, ByVal max As Vector, ByRef p As Vector) As Boolean
        If (code And Edge.LEFT) Then
            If IntsecX(p0, p1, min.x(0), min, max, p) Then
                Return True
            End If
        End If
        If (code And Edge.RIGHT) Then
            If IntsecX(p0, p1, max.x(0), min, max, p) Then
                Return True
            End If
        End If
        If (code And Edge.TOP) Then
            If IntsecY(p0, p1, min.x(1), min, max, p) Then
                Return True
            End If
        End If
        If (code And Edge.BOTTOM) Then
            If IntsecY(p0, p1, max.x(1), min, max, p) Then
                Return True
            End If
        End If
        Return False
    End Function

    '線分クリップ
    Public Function LineClipping(ByRef p0 As Vector, ByRef p1 As Vector, ByVal min As Vector, ByVal max As Vector) As Integer
        Dim code0 As Edge
        Dim code1 As Edge
        Dim v0 As Vector = p0.Clone()
        Dim v1 As Vector = p1.Clone()
        code0 = SeqCode(p0, min, max)
        code1 = SeqCode(p1, min, max)
        If (code0 = 0) And (code1 = 0) Then
            Return 0
        End If
        If (code0 And code1) Then
            Return -1
        End If
        If code0 Then
            If Not LineClippedPoint(code0, p0, p1, min, max, v0) Then
                Return -1
            End If
        End If
        If code1 Then
            If Not LineClippedPoint(code1, p0, p1, min, max, v1) Then
                Return -1
            End If
        End If
        Dim bx As Double = (p1.x(2) - p0.x(2)) / (p1.x(0) - p0.x(0))
        If Double.IsNaN(bx) Or Double.IsInfinity(bx) Then
            bx = p1.x(2) - p0.x(2)
        End If
        Dim dx0 As Double = p0.x(0) - v0.x(0)
        Dim dx1 As Double = p1.x(0) - v1.x(0)
        v0.x(2) = dx0 * bx + p0.x(2)
        v1.x(2) = dx1 * bx + p1.x(2)
        p0 = v0
        p1 = v1
        Return 1
    End Function


    '直線描画
    Public Sub DrawLine(ByRef SB As SpriteBatch, ByVal white As Texture2D, ByVal vcol As Color, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal p1 As Point, ByVal p2 As Point)
        Dim ol As Integer = l - w / 2
        Dim ot As Integer = t - h / 2

        Dim dx As Integer = p2.X - p1.X
        Dim dy As Integer = p2.Y - p1.Y
        Dim adx As Integer = Math.Abs(dx)
        Dim ady As Integer = Math.Abs(dy)

        If (ady < adx And dx < 0) Or (adx <= ady And dy < 0) Then
            Dim pWK As New Point(p1.X, p1.Y)
            p1 = p2
            p2 = pWK
            dx = -dx
            dy = -dy
        End If

        Dim x As Integer
        Dim y As Integer
        Dim e As Integer = 0



        If ady < adx Then
            y = p1.Y
            For x = p1.X To p2.X
                e += ady
                If adx < e Then
                    e -= adx
                    If dy < 0 Then
                        y -= 1
                    Else
                        y += 1
                    End If
                End If
                Dim pp As Point = New Point(x, y)
                If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                    SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), vcol)
                End If
            Next
        Else
            x = p1.X
            For y = p1.Y To p2.Y
                e += adx
                If ady < e Then
                    e -= ady
                    If dx < 0 Then
                        x -= 1
                    Else
                        x += 1
                    End If
                End If
                Dim pp As Point = New Point(x, y)
                If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                    SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), vcol)
                End If
            Next
        End If
    End Sub

    '直線描画
    Public Sub DrawLine(ByRef SB As SpriteBatch, ByVal white As Texture2D, ByVal vcol As Color, ByVal n As Vector, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal p1 As Point, ByVal p2 As Point, ByVal z1 As Double, ByVal z2 As Double, ByVal axz As Double, ByVal ayz As Double, ByVal mtr As Material, ByVal cmrmat As Matrix)
        Dim InfWK As New Infinity(Range, Base)
        Dim ol As Integer = l - w / 2
        Dim ot As Integer = t - h / 2

        Dim dx As Integer = p2.X - p1.X
        Dim dy As Integer = p2.Y - p1.Y
        Dim adx As Integer = Math.Abs(dx)
        Dim ady As Integer = Math.Abs(dy)

        If (ady < adx And dx < 0) Or (adx <= ady And dy < 0) Then
            Dim pWK As New Point(p1.X, p1.Y)
            p1 = p2
            p2 = pWK
            dx = -dx
            dy = -dy
            Dim tmp As Double = z1
            z1 = z2
            z2 = tmp
            axz = -axz
            ayz = -ayz
        End If

        Dim x As Integer
        Dim y As Integer
        Dim z As Double = z1
        Dim e As Integer = 0



        If ady < adx Then
            y = p1.Y
            For x = p1.X To p2.X
                e += ady
                If adx < e Then
                    e -= adx
                    If dy < 0 Then
                        y -= 1
                    Else
                        y += 1
                    End If
                End If
                Dim pp As Point = New Point(x, y)
                If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                    'ｚオーダークリップ
                    Dim zb As Double = ZBuf(pp.Y - ot, pp.X - ol)
                    If (z < zb Or zb = -1.0) Then ' And Math.Abs(LineDelta / z1) < z Then
                        Dim pv As New Vector(4, True)
                        pv.x(0) = 1.0
                        pv.x(1) = (CType(pp.X, Double) - l) / CType(w / 2, Double)
                        pv.x(2) = -(CType(pp.Y, Double) - t) / CType(h / 2, Double)
                        pv.x(3) = 0.0
                        If BInf Then
                            pv = InfWK.InvFrustumInf(pv, z)
                        Else
                            pv.x(3) = z
                        End If
                        Dim col As Color = lgt.GetColor(vcol, n, pv, mtr, cmrmat, False)
                        SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), col)
                        ZBuf(pp.Y - ot, pp.X - ol) = z - Math.Abs(LineDelta / z1)
                    End If
                End If
                z += axz
            Next
        Else
            x = p1.X
            For y = p1.Y To p2.Y
                e += adx
                If ady < e Then
                    e -= ady
                    If dx < 0 Then
                        x -= 1
                    Else
                        x += 1
                    End If
                End If
                Dim pp As Point = New Point(x, y)
                If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                    'ｚオーダークリップ
                    Dim zb As Double = ZBuf(pp.Y - ot, pp.X - ol)
                    If (z < zb Or zb = -1.0) Then 'And Math.Abs(LineDelta / z1) < z Then
                        Dim pv As New Vector(4, True)
                        pv.x(0) = 1.0
                        pv.x(1) = (CType(pp.X, Double) - l) / CType(w / 2, Double)
                        pv.x(2) = -(CType(pp.Y, Double) - t) / CType(h / 2, Double)
                        pv.x(3) = 0.0
                        If BInf Then
                            pv = InfWK.InvFrustumInf(pv, z)
                        Else
                            pv.x(3) = z
                        End If
                        Dim col As Color = lgt.GetColor(vcol, n, pv, mtr, cmrmat, False)
                        SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), col)
                        ZBuf(pp.Y - ot, pp.X - ol) = z - Math.Abs(LineDelta / z1)
                    End If
                End If
                z += ayz
            Next
        End If
    End Sub

    '線分クリップ描画
    Public Sub DrawClippingLine(ByRef SB As SpriteBatch, ByVal white As Texture2D, ByVal vcol As Color, ByVal n As Vector, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal v1 As Vector, ByVal v2 As Vector, ByVal mtr As Material, ByVal cmrmat As Matrix)
        'If v1.x(3) < 0.0 Or v2.x(3) < 0.0 Then
        'Return
        'End If

        Dim InfWK As New Infinity(Range, Base)
        Dim VecWK1 As Vector = v1.Clone()
        Dim VecWK2 As Vector = v2.Clone()
        VecWK1 = VecWK1.Homogeneous()
        VecWK2 = VecWK2.Homogeneous()

        Dim VecWK1F As Vector
        Dim VecWK2F As Vector
        '透視射影
        If BInf Then
            VecWK1F = InfWK.FrustumInf(VecWK1)
            VecWK2F = InfWK.FrustumInf(VecWK2)
        Else
            VecWK1F = Frs.Frustum(VecWK1)
            VecWK2F = Frs.Frustum(VecWK2)
            VecWK1F.x(0) = 1.0
            VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
            VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
            VecWK1F.x(3) = VecWK1.x(3)
            VecWK2F.x(0) = 1.0
            VecWK2F.x(1) = -VecWK2F.x(1) / VecWK2F.x(3)
            VecWK2F.x(2) = -VecWK2F.x(2) / VecWK2F.x(3)
            VecWK2F.x(3) = VecWK2.x(3)
        End If

        Dim min As New Vector(2)
        Dim max As New Vector(2)
        Dim clip1 As New Vector(3)
        Dim clip2 As New Vector(3)

        min.x(0) = CType(-w / 2, Double) + CType(l, Double)
        min.x(1) = CType(-h / 2, Double) + CType(t, Double)
        max.x(0) = CType(w / 2, Double) + CType(l, Double)
        max.x(1) = CType(h / 2, Double) + CType(t, Double)


        clip1.x(0) = VecWK1F.x(1) * CType(w / 2, Double) + CType(l, Double)
        clip1.x(1) = -VecWK1F.x(2) * CType(h / 2, Double) + CType(t, Double)
        clip1.x(2) = VecWK1.x(3)
        clip2.x(0) = VecWK2F.x(1) * CType(w / 2, Double) + CType(l, Double)
        clip2.x(1) = -VecWK2F.x(2) * CType(h / 2, Double) + CType(t, Double)
        clip2.x(2) = VecWK2.x(3)

        If 0 <= LineClipping(clip1, clip2, min, max) Then
            Dim minz As Vector = min
            Dim maxz As Vector = max
            minz.x(1) = 0.0
            maxz.x(1) = Double.MaxValue
            Dim clz1 As New Vector(3)
            Dim clz2 As New Vector(3)
            clz1.x(0) = clip1.x(0)
            clz1.x(1) = clip1.x(2)
            clz1.x(2) = clip1.x(1)
            clz2.x(0) = clip2.x(0)
            clz2.x(1) = clip2.x(2)
            clz2.x(2) = clip2.x(1)
            If 0 <= LineClipping(clz1, clz2, minz, maxz) Then
                clip1.x(0) = clz1.x(0)
                clip1.x(1) = clz1.x(2)
                clip1.x(2) = clz1.x(1)
                clip2.x(0) = clz2.x(0)
                clip2.x(1) = clz2.x(2)
                clip2.x(2) = clz2.x(1)

                Dim p1 As New Point(CType(clip1.x(0), Integer), CType(clip1.x(1), Integer))
                Dim p2 As New Point(CType(clip2.x(0), Integer), CType(clip2.x(1), Integer))
                Dim z1 As Double = clip1.x(2)
                Dim z2 As Double = clip2.x(2)
                Dim bx1 As Double = (clip2.x(2) - clip1.x(2)) / (clip2.x(0) - clip1.x(0))
                Dim by1 As Double = (clip2.x(2) - clip1.x(2)) / (clip2.x(1) - clip1.x(1))

                If Double.IsNaN(bx1) Or Double.IsInfinity(bx1) Then
                    bx1 = (clip2.x(2) - clip1.x(2))
                End If
                If Double.IsNaN(by1) Or Double.IsInfinity(by1) Then
                    by1 = (clip2.x(2) - clip1.x(2))
                End If


                DrawLine(SB, white, vcol, n, w, h, l, t, p1, p2, z1, z2, bx1, by1, mtr, cmrmat)
            End If
        End If
    End Sub

    '三角形描画
    Public Sub DrawTriangle(ByRef SB As SpriteBatch, ByVal white As Texture2D, ByVal vcol As Color, ByVal n As Vector, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal v1 As Vector, ByVal v2 As Vector, ByVal v3 As Vector, ByVal mtr As Material, ByVal cmrmat As Matrix)
        DrawClippingLine(SB, white, vcol, n, w, h, l, t, v1, v2, mtr, cmrmat)
        DrawClippingLine(SB, white, vcol, n, w, h, l, t, v2, v3, mtr, cmrmat)
        DrawClippingLine(SB, white, vcol, n, w, h, l, t, v3, v1, mtr, cmrmat)
    End Sub

    '三角形塗りつぶし
    Public Sub FillTriangle(ByRef SB As SpriteBatch, ByVal white As Texture2D, ByVal tcol As Color, ByVal n As Vector, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal v1 As Vector, ByVal v2 As Vector, ByVal v3 As Vector, ByVal mtr As Material, ByVal cmrmat As Matrix)
        'If v1.x(3) < 0.0 Or v2.x(3) < 0.0 Or v3.x(3) < 0.0 Then
        'Return
        'End If
        Dim InfWK As New Infinity(Range, Base)
        Dim VecWK1 As Vector = v1.Clone()
        Dim VecWK2 As Vector = v2.Clone()
        Dim VecWK3 As Vector = v3.Clone()
        VecWK1 = VecWK1.Homogeneous()
        VecWK2 = VecWK2.Homogeneous()
        VecWK3 = VecWK3.Homogeneous()

        Dim VecWK1F As Vector
        Dim VecWK2F As Vector
        Dim VecWK3F As Vector
        '透視射影
        If BInf Then
            VecWK1F = InfWK.FrustumInf(VecWK1)
            VecWK2F = InfWK.FrustumInf(VecWK2)
            VecWK3F = InfWK.FrustumInf(VecWK3)
        Else
            VecWK1F = Frs.Frustum(VecWK1)
            VecWK2F = Frs.Frustum(VecWK2)
            VecWK3F = Frs.Frustum(VecWK3)
            VecWK1F.x(0) = 1.0
            VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
            VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
            VecWK1F.x(3) = VecWK1.x(3)
            VecWK2F.x(0) = 1.0
            VecWK2F.x(1) = -VecWK2F.x(1) / VecWK2F.x(3)
            VecWK2F.x(2) = -VecWK2F.x(2) / VecWK2F.x(3)
            VecWK2F.x(3) = VecWK2.x(3)
            VecWK3F.x(0) = 1.0
            VecWK3F.x(1) = -VecWK3F.x(1) / VecWK3F.x(3)
            VecWK3F.x(2) = -VecWK3F.x(2) / VecWK3F.x(3)
            VecWK3F.x(3) = VecWK3.x(3)
        End If

        Dim p1 As New Vector(2)
        p1.x(0) = CType(VecWK1F.x(1) * CType(w / 2, Double), Integer) + l
        p1.x(1) = CType(-VecWK1F.x(2) * CType(h / 2, Double), Integer) + t
        Dim p2 As New Vector(2)
        p2.x(0) = CType(VecWK2F.x(1) * CType(w / 2, Double), Integer) + l
        p2.x(1) = CType(-VecWK2F.x(2) * CType(h / 2, Double), Integer) + t
        Dim p3 As New Vector(2)
        p3.x(0) = CType(VecWK3F.x(1) * CType(w / 2, Double), Integer) + l
        p3.x(1) = CType(-VecWK3F.x(2) * CType(h / 2, Double), Integer) + t

        Dim ol As Integer = l - w / 2
        Dim ot As Integer = t - h / 2

        If p2.x(1) < p1.x(1) And p2.x(1) <= p3.x(1) Then
            Dim VecWK As Vector = VecWK1.Clone()
            VecWK1 = VecWK2.Clone()
            VecWK2 = VecWK.Clone()
            VecWK = VecWK1F.Clone()
            VecWK1F = VecWK2F.Clone()
            VecWK2F = VecWK.Clone()
            VecWK = p1.Clone()
            p1 = p2.Clone()
            p2 = VecWK.Clone()
        Else
            If p3.x(1) < p1.x(1) And p3.x(1) <= p2.x(1) Then
                Dim VecWK As Vector = VecWK1.Clone()
                VecWK1 = VecWK3.Clone()
                VecWK3 = VecWK.Clone()
                VecWK = VecWK1F.Clone()
                VecWK1F = VecWK3F.Clone()
                VecWK3F = VecWK.Clone()
                VecWK = p1.Clone()
                p1 = p3.Clone()
                p3 = VecWK.Clone()
            End If
        End If

        If p3.x(1) < p2.x(1) Then
            Dim VecWK As Vector = VecWK2.Clone()
            VecWK2 = VecWK3.Clone()
            VecWK3 = VecWK.Clone()
            VecWK = VecWK2F.Clone()
            VecWK2F = VecWK3F.Clone()
            VecWK3F = VecWK.Clone()
            VecWK = p2.Clone()
            p2 = p3.Clone()
            p3 = VecWK.Clone()
        End If

        Dim a1 As Double = (p2.x(0) - p1.x(0)) / (p2.x(1) - p1.x(1))
        Dim a2 As Double = (p3.x(0) - p1.x(0)) / (p3.x(1) - p1.x(1))
        Dim a3 As Double = (p3.x(0) - p2.x(0)) / (p3.x(1) - p2.x(1))

        Dim bz1 As Double = (VecWK2.x(3) - VecWK1.x(3)) / (p2.x(1) - p1.x(1))
        Dim bz2 As Double = (VecWK3.x(3) - VecWK1.x(3)) / (p3.x(1) - p1.x(1))
        Dim bz3 As Double = (VecWK3.x(3) - VecWK2.x(3)) / (p3.x(1) - p2.x(1))

        If Double.IsNaN(a1) Or Double.IsInfinity(a1) Then
            a1 = (p2.x(0) - p1.x(0))
        End If
        If Double.IsNaN(a2) Or Double.IsInfinity(a2) Then
            a2 = (p3.x(0) - p1.x(0))
        End If
        If Double.IsNaN(a3) Or Double.IsInfinity(a3) Then
            a3 = (p3.x(0) - p2.x(0))
        End If
        If Double.IsNaN(bz1) Or Double.IsInfinity(bz1) Then
            bz1 = (VecWK2.x(3) - VecWK1.x(3))
        End If
        If Double.IsNaN(bz2) Or Double.IsInfinity(bz2) Then
            bz2 = (VecWK3.x(3) - VecWK1.x(3))
        End If
        If Double.IsNaN(bz3) Or Double.IsInfinity(bz3) Then
            bz3 = (VecWK3.x(3) - VecWK2.x(3))
        End If

        Dim px1 As Double = p1.x(0)
        Dim px2 As Double = p1.x(0)
        Dim x As Integer
        Dim y As Integer

        Dim ssx As Integer = p1.x(0)
        Dim eex As Integer = p1.x(0)
        If p2.x(0) < ssx Then
            ssx = p2.x(0)
        End If
        If p3.x(0) < ssx Then
            ssx = p3.x(0)
        End If
        If eex < p2.x(0) Then
            eex = p2.x(0)
        End If
        If eex < p3.x(0) Then
            eex = p3.x(0)
        End If



        Dim ppz1 As Double = VecWK1.x(3)
        Dim ppz2 As Double = VecWK1.x(3)

        Dim cymin As Integer = p1.x(1)
        Dim cymax As Integer = p3.x(1) - 1
        If cymin < -h / 2 + t Then
            Dim cy As Integer = -h / 2 + t
            Dim dy As Integer = cy - cymin
            Dim xxx As Double
            Dim zzz As Double
            If p2.x(1) < cy Then
                px1 = p2.x(0)
                ppz1 = VecWK2.x(3)
                xxx = a3
                zzz = bz3
                dy = cy - p2.x(1)
            Else
                xxx = a1
                zzz = bz1
            End If
            px1 += xxx * dy
            ppz1 += zzz * dy
            px2 += a2 * dy
            ppz2 += bz2 * dy
        End If
        If h / 2 + t < cymax Then
            cymax = h / 2 + t
        End If

        For y = cymin To cymax
            If y = p2.x(1) Then
                px1 = p2.x(0)
                ppz1 = VecWK2.x(3)
            ElseIf y < p2.x(1) Then
                px1 += a1
                ppz1 += bz1
            Else
                px1 += a3
                ppz1 += bz3
            End If
            px2 += a2
            ppz2 += bz2

            Dim sx As Integer
            Dim ex As Integer
            Dim psz As Double
            Dim pez As Double
            If px1 < px2 Then
                sx = px1
                ex = px2
                psz = ppz1
                pez = ppz2
            Else
                sx = px2
                ex = px1
                psz = ppz2
                pez = ppz1
            End If
            If sx < ssx Then
                sx = ssx
            End If
            If eex < ex Then
                ex = eex
            End If
            If sx < -w / 2 + l Then
                sx = -w / 2 + l
            End If
            If w / 2 + l < ex Then
                ex = w / 2 + l
            End If
            Dim vzz As Double = (pez - psz) / (CType(ex - sx, Double))
            Dim vz As Double = psz
            For x = sx To ex
                Dim pp As New Point(x, y)
                If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                    'ｚオーダークリップ
                    Dim zb As Double = ZBuf(pp.Y - ot, pp.X - ol)
                    If (vz < zb Or zb = -1.0) And 0.0 < vz Then
                        Dim pv As New Vector(4, True)
                        pv.x(0) = 1.0
                        pv.x(1) = (CType(pp.X, Double) - l) / CType(w / 2, Double)
                        pv.x(2) = -(CType(pp.Y, Double) - t) / CType(h / 2, Double)
                        pv.x(3) = 0.0
                        If BInf Then
                            pv = InfWK.InvFrustumInf(pv, vz)
                        Else
                            pv.x(3) = vz
                        End If
                        Dim col As Color = lgt.GetColor(tcol, n, pv, mtr, cmrmat, False)
                        SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), col)
                        ZBuf(pp.Y - ot, pp.X - ol) = vz
                    End If
                End If
                vz += vzz
            Next
        Next
    End Sub

    'zと法線の取得
    Public Sub SetZN(ByRef z As Double, ByRef n As Vector, ByVal v1 As Vector, ByVal v2 As Vector, ByVal r As Double, ByVal x As Integer, ByVal y As Integer, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer)
        Dim InfWK As New Infinity(Range, Base)
        Dim VecWK1 As Vector = v1.Clone()
        Dim VecWK2 As Vector = v2.Clone()
        VecWK1 = VecWK1.Homogeneous()
        VecWK2 = VecWK2.Homogeneous()

        Dim VecWK1F As Vector
        Dim VecWK2F As Vector
        '透視射影
        If BInf Then
            VecWK1F = InfWK.FrustumInf(VecWK1)
            VecWK2F = InfWK.FrustumInf(VecWK2)
        Else
            VecWK1F = Frs.Frustum(VecWK1)
            VecWK2F = Frs.Frustum(VecWK2)
            VecWK1F.x(0) = 1.0
            VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
            VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
            VecWK1F.x(3) = VecWK1.x(3)
            VecWK2F.x(0) = 1.0
            VecWK2F.x(1) = -VecWK2F.x(1) / VecWK2F.x(3)
            VecWK2F.x(2) = -VecWK2F.x(2) / VecWK2F.x(3)
            VecWK2F.x(3) = VecWK2.x(3)
        End If

        Dim rxx As Integer = CType(VecWK1F.x(1) * CType(w / 2, Double), Integer) + l
        Dim ryy As Integer = CType(-VecWK1F.x(2) * CType(h / 2, Double), Integer) + t
        Dim rww As Integer = CType(Math.Abs(CType(VecWK2F.x(1) * CType(w / 2, Double), Integer) + l - rxx), Integer)
        Dim rhh As Integer = CType(Math.Abs(CType(-VecWK2F.x(2) * CType(h / 2, Double), Integer) + l - ryy), Integer)

        Dim rect As New Rectangle(rxx, ryy, rww, rhh)


        Dim ww As Integer = rect.Width / 2 + 1
        Dim hh As Integer = rect.Height / 2 + 1

        Dim th As Double = Math.Acos(x / ww)
        If Double.IsNaN(th) Then
            th = 0.0
        End If
        Dim thz As Double = Math.Acos(y / hh)
        Dim nx As New Vector(4, True)
        nx = nx.UnitVec(1) * r

        Dim nz As New Vector(4, True)
        nz = nz.UnitVec(3) * -r

        nz = nz.RotAxisQuat(nx, thz)
        n = nx.RotAxisQuat(nz, th)
        If Math.Abs(th) = Math.PI Then
            z = nz.x(3) + VecWK1.x(3)
        Else
            z = n.x(3) + VecWK1.x(3)
        End If
        n = n.NormalVec()
    End Sub


    '楕円形描画
    Public Sub DrawEllipse(ByRef SB As SpriteBatch, ByVal white As Texture2D, ByVal vcol As Color, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal v1 As Vector, ByVal v2 As Vector, ByVal r As Double, ByVal mtr As Material, ByVal cmrmat As Matrix)
        If v1.x(3) < 0.0 Or v2.x(3) < 0.0 Then
            Return
        End If
        Dim InfWK As New Infinity(Range, Base)
        Dim VecWK1 As Vector = v1.Clone()
        Dim VecWK2 As Vector = v2.Clone()
        VecWK1 = VecWK1.Homogeneous()
        VecWK2 = VecWK2.Homogeneous()

        Dim VecWK1F As Vector
        Dim VecWK2F As Vector
        '透視射影
        If BInf Then
            VecWK1F = InfWK.FrustumInf(VecWK1)
            VecWK2F = InfWK.FrustumInf(VecWK2)
        Else
            VecWK1F = Frs.Frustum(VecWK1)
            VecWK2F = Frs.Frustum(VecWK2)
            VecWK1F.x(0) = 1.0
            VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
            VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
            VecWK1F.x(3) = VecWK1.x(3)
            VecWK2F.x(0) = 1.0
            VecWK2F.x(1) = -VecWK2F.x(1) / VecWK2F.x(3)
            VecWK2F.x(2) = -VecWK2F.x(2) / VecWK2F.x(3)
            VecWK2F.x(3) = VecWK2.x(3)
        End If

        Dim rxx As Integer = CType(VecWK1F.x(1) * CType(w / 2, Double), Integer) + l
        Dim ryy As Integer = CType(-VecWK1F.x(2) * CType(h / 2, Double), Integer) + t
        Dim rww As Integer = CType(Math.Abs(CType(VecWK2F.x(1) * CType(w / 2, Double), Integer) + l - rxx), Integer)
        Dim rhh As Integer = CType(Math.Abs(CType(-VecWK2F.x(2) * CType(h / 2, Double), Integer) + l - ryy), Integer)

        Dim rect As New Rectangle(rxx, ryy, rww, rhh)

        Dim ol As Integer = l - w / 2
        Dim ot As Integer = t - h / 2


        Dim i As Integer
        Dim x As Integer
        Dim y As Integer
        Dim z As Double
        Dim n As New Vector(4, True)


        Dim ww As Integer = rect.Width / 2 + 1
        Dim hh As Integer = rect.Height / 2 + 1

        Dim cy As Integer = hh
        If (h / 2 + t) / 2 + 1 < cy Then
            cy = (h / 2 + t) / 2 + 1
        End If



        For y = -cy To -1
            Dim theta As Double = Math.Asin(y / Math.Abs(hh))
            Dim xx As Double = Math.Cos(theta) * Math.Abs(ww)
            If (w / 2 + l) / 2 + 1 < Math.Abs(xx) Then
                Continue For
            End If
            Dim theta2 As Double = Math.Asin((y + 1) / Math.Abs(hh))
            Dim xx2 As Double = Math.Cos(theta2) * Math.Abs(ww)
            Dim pp As Point
            Dim yy As Integer = y
            For i = 0 To 1
                If i = 1 Then
                    yy = -yy - 1
                End If
                For x = -xx2 To -xx
                    pp = New Point(x + rect.X + ww, yy + rect.Y + hh)
                    If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                        'ｚオーダークリップ
                        Dim zb As Double = ZBuf(pp.Y - ot, pp.X - ol)
                        SetZN(z, n, v1, v2, r, x, yy, w, h, l, t)
                        If (z < zb Or zb = -1.0) Then 'And Math.Abs(LineDelta / z) < z Then
                            Dim pv As New Vector(4, True)
                            pv.x(0) = 1.0
                            pv.x(1) = (CType(pp.X, Double) - l) / CType(w / 2, Double)
                            pv.x(2) = -(CType(pp.Y, Double) - t) / CType(h / 2, Double)
                            pv.x(3) = 0.0
                            If BInf Then
                                pv = InfWK.InvFrustumInf(pv, z)
                            Else
                                pv.x(3) = z
                            End If
                            Dim col As Color = lgt.GetColor(vcol, n, pv, mtr, cmrmat, True)
                            SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), col)
                            ZBuf(pp.Y - ot, pp.X - ol) = z - Math.Abs(LineDelta / z)
                        End If
                    End If
                Next
                For x = xx To xx2
                    pp = New Point(x + rect.X + ww, yy + rect.Y + hh)
                    If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                        'ｚオーダークリップ
                        Dim zb As Double = ZBuf(pp.Y - ot, pp.X - ol)
                        SetZN(z, n, v1, v2, r, x, yy, w, h, l, t)
                        If (z < zb Or zb = -1.0) Then 'And Math.Abs(LineDelta / z) < z Then
                            Dim pv As New Vector(4, True)
                            pv.x(0) = 1.0
                            pv.x(1) = (CType(pp.X, Double) - l) / CType(w / 2, Double)
                            pv.x(2) = -(CType(pp.Y, Double) - t) / CType(h / 2, Double)
                            pv.x(3) = 0.0
                            If BInf Then
                                pv = InfWK.InvFrustumInf(pv, z)
                            Else
                                pv.x(3) = z
                            End If
                            Dim col As Color = lgt.GetColor(vcol, n, pv, mtr, cmrmat, True)
                            SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), col)
                            ZBuf(pp.Y - ot, pp.X - ol) = z - Math.Abs(LineDelta / z)
                        End If
                    End If
                Next
            Next
        Next
    End Sub

    '楕円形塗りつぶし
    Public Sub FillEllipse(ByRef SB As SpriteBatch, ByVal white As Texture2D, ByVal vcol As Color, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal v1 As Vector, ByVal v2 As Vector, ByVal r As Double, ByVal mtr As Material, ByVal cmrmat As Matrix)
        If v1.x(3) < 0.0 Or v2.x(3) < 0.0 Then
            Return
        End If
        Dim InfWK As New Infinity(Range, Base)
        Dim VecWK1 As Vector = v1.Clone()
        Dim VecWK2 As Vector = v2.Clone()
        VecWK1 = VecWK1.Homogeneous()
        VecWK2 = VecWK2.Homogeneous()

        Dim VecWK1F As Vector
        Dim VecWK2F As Vector
        '透視射影
        If BInf Then
            VecWK1F = InfWK.FrustumInf(VecWK1)
            VecWK2F = InfWK.FrustumInf(VecWK2)
        Else
            VecWK1F = Frs.Frustum(VecWK1)
            VecWK2F = Frs.Frustum(VecWK2)
            VecWK1F.x(0) = 1.0
            VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
            VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
            VecWK1F.x(3) = VecWK1.x(3)
            VecWK2F.x(0) = 1.0
            VecWK2F.x(1) = -VecWK2F.x(1) / VecWK2F.x(3)
            VecWK2F.x(2) = -VecWK2F.x(2) / VecWK2F.x(3)
            VecWK2F.x(3) = VecWK2.x(3)
        End If

        Dim rxx As Integer = CType(VecWK1F.x(1) * CType(w / 2, Double), Integer) + l
        Dim ryy As Integer = CType(-VecWK1F.x(2) * CType(h / 2, Double), Integer) + t
        Dim rww As Integer = CType(Math.Abs(CType(VecWK2F.x(1) * CType(w / 2, Double), Integer) + l - rxx), Integer)
        Dim rhh As Integer = CType(Math.Abs(CType(-VecWK2F.x(2) * CType(h / 2, Double), Integer) + l - ryy), Integer)

        Dim rect As New Rectangle(rxx, ryy, rww, rhh)

        Dim ol As Integer = l - w / 2
        Dim ot As Integer = t - h / 2


        Dim i As Integer
        Dim x As Integer
        Dim y As Integer
        Dim z As Double
        Dim n As New Vector(4, True)


        Dim ww As Integer = rect.Width / 2 + 1
        Dim hh As Integer = rect.Height / 2 + 1

        Dim cx As Integer
        Dim cy As Integer = hh
        If (h / 2 + t) / 2 + 1 < cy Then
            cy = (h / 2 + t) / 2 + 1
        End If



        For y = -cy To -1
            Dim theta As Double = Math.Asin(y / Math.Abs(hh))
            Dim xx As Double = Math.Cos(theta) * Math.Abs(ww)
            Dim theta2 As Double = Math.Asin((y + 1) / Math.Abs(hh))
            Dim xx2 As Double = Math.Cos(theta2) * Math.Abs(ww)
            cx = Math.Abs(xx2)
            If (w / 2 + l) / 2 + 1 < cx Then
                cx = (w / 2 + l) / 2 + 1
            End If

            Dim pp As Point
            Dim yy As Integer = y
            For i = 0 To 1
                If i = 1 Then
                    yy = -yy - 1
                End If
                For x = -cx To cx
                    pp = New Point(x + rect.X + ww, yy + rect.Y + hh)
                    If 0 <= pp.X - ol And pp.X - ol <= w And 0 <= pp.Y - ot And pp.Y - ot <= h Then
                        'ｚオーダークリップ
                        Dim zb As Double = ZBuf(pp.Y - ot, pp.X - ol)
                        SetZN(z, n, v1, v2, r, x, yy, w, h, l, t)
                        If (z < zb Or zb = -1.0) Then 'And 0.0 < z Then
                            Dim pv As New Vector(4, True)
                            pv.x(0) = 1.0
                            pv.x(1) = (CType(pp.X, Double) - l) / CType(w / 2, Double)
                            pv.x(2) = -(CType(pp.Y, Double) - t) / CType(h / 2, Double)
                            pv.x(3) = 0.0
                            If BInf Then
                                pv = InfWK.InvFrustumInf(pv, z)
                            Else
                                pv.x(3) = z
                            End If
                            Dim col As Color = lgt.GetColor(vcol, n, pv, mtr, cmrmat, True)
                            SB.Draw(white, New Rectangle(pp.X, pp.Y, 1, 1), col)
                            ZBuf(pp.Y - ot, pp.X - ol) = z
                        End If
                    End If
                Next
            Next
        Next
    End Sub

    'ラベル描画チェック
    Public Function CheckLabel(ByVal v1 As Vector, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer) As Boolean
        If v1.x(3) < 0.0 Then
            Return False
        End If
        Dim InfWK As New Infinity(Range, Base)
        Dim VecWK1 As Vector = v1.Clone()
        VecWK1 = VecWK1.Homogeneous()

        Dim VecWK1F As Vector
         If BInf Then
            VecWK1F = InfWK.FrustumInf(VecWK1)
        Else
            VecWK1F = Frs.Frustum(VecWK1)
            VecWK1F.x(0) = 1.0
            VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
            VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
            VecWK1F.x(3) = VecWK1.x(3)
       End If

        Dim xxx As Double = VecWK1F.x(1) * CType(w / 2, Double) + CType(l, Double)
        Dim yyy As Double = -VecWK1F.x(2) * CType(h / 2, Double) + CType(t, Double)


        Dim ol As Integer = l - w / 2
        Dim ot As Integer = t - h / 2
        If 0.0 <= xxx - CType(ol, Double) And xxx - CType(ol, Double) <= CType(w, Double) And 0.0 <= yyy - CType(ot, Double) And yyy - CType(ot, Double) <= CType(h, Double) Then
            Dim p1 As New Point(xxx, yyy)

            'ｚオーダークリップ
            Dim z As Double = VecWK1.x(3)
            Dim zb As Double = ZBuf(p1.Y - ot, p1.X - ol)
            If (z < zb Or zb = -1.0) Then 'And 0.0 < z Then
                Return True
            End If
        End If
        Return False
    End Function


    'ラベル描画
    Public Sub DrawLabel(ByRef SB As SpriteBatch, ByRef white As Texture2D, ByVal v1 As Vector, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal lbl As Label)
        If lbl.str.Text = "" Then
            Return
        End If
        Dim InfWK As New Infinity(Range, Base)
        Dim VecWK1 As Vector = v1.Clone()
        VecWK1 = VecWK1.Homogeneous()

        Dim VecWK1F As Vector
        If BInf Then
            VecWK1F = InfWK.FrustumInf(VecWK1)
        Else
            VecWK1F = Frs.Frustum(VecWK1)
            VecWK1F.x(0) = 1.0
            VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
            VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
            VecWK1F.x(3) = VecWK1.x(3)
       End If

        Dim xxx As Double = VecWK1F.x(1) * CType(w / 2, Double) + CType(l, Double)
        Dim yyy As Double = -VecWK1F.x(2) * CType(h / 2, Double) + CType(t, Double)


        Dim ol As Integer = l - w / 2
        Dim ot As Integer = t - h / 2
        If 0.0 <= xxx - CType(ol, Double) And xxx - CType(ol, Double) <= CType(w, Double) And 0.0 <= yyy - CType(ot, Double) And yyy - CType(ot, Double) <= CType(h, Double) Then
            Dim p1 As New Point(xxx, yyy)

            Dim scl As New Vector2(0.4, 0.4)
            Dim z As Double = VecWK1.x(3)
            Dim x1 As Integer
            Dim x2 As Integer
            Dim y1 As Integer
            Dim x As Integer
            Dim y As Integer
            If lbl.offset.X < 0 Then
                x = p1.X + lbl.offset.X - lbl.str.Size.X * scl.X
                y = p1.Y + lbl.offset.Y
                x1 = x + lbl.str.Size.X * scl.X
                x2 = x
                y1 = y + lbl.str.Size.Y * scl.Y
            Else
                x = p1.X + lbl.offset.X
                y = p1.Y + lbl.offset.Y
                x1 = x
                x2 = x + lbl.str.Size.X * scl.X
                y1 = y + lbl.str.Size.Y * scl.Y
            End If
            DrawLine(SB, white, lbl.col, w, h, l, t, p1, New Point(x1, y1))
            DrawLine(SB, white, lbl.col, w, h, l, t, New Point(x1, y1), New Point(x2, y1))
            'SB.DrawString(lbl.str.Font, lbl.str.Text, New Vector2(x, y), lbl.col)
            SB.DrawString(lbl.str.Font, lbl.str.Text, New Vector2(x, y), lbl.col, 0.0, New Vector2(0, 0), scl, SpriteEffects.None, 0)
        End If
    End Sub

    '描画
    'gPic:グラフィックスオブジェクト,w:画面幅,h:画面高さ,l:左オフセット,t:上オフセット,b:true:通常描画,false:ワイヤーフレーム陰面消去なし
    Public Sub Render(ByRef SB As SpriteBatch, ByRef white As Texture2D, ByVal w As Integer, ByVal h As Integer, ByVal l As Integer, ByVal t As Integer, ByVal b As Boolean)
        ReDim ZBuf(h, w)
        Dim i As Integer
        Dim x As Integer
        Dim y As Integer
        For y = 0 To h
            For x = 0 To w
                ZBuf(y, x) = -1
            Next
        Next

        Dim pvwk(Poly.Count - 1) As Vector
        Dim pbwk(Poly.Count - 1) As Boolean
        Dim svwk(Sphr.Count - 1) As Vector
        Dim sbwk(Sphr.Count - 1) As Boolean
        Dim lvwk(Line.Count - 1) As Vector
        Dim lbwk(Line.Count - 1) As Boolean
        For i = 0 To Poly.Count - 1
            pbwk(i) = False
        Next
        For i = 0 To Sphr.Count - 1
            sbwk(i) = False
        Next
        For i = 0 To Line.Count - 1
            lbwk(i) = False
        Next



        lgt.Mat = lgt.px.Mat()


        'ポリゴン描画
        For i = 0 To Poly.Count - 1
            'オブジェクトローカル座標系
            Dim dt As Double
            Dim IntWK As Integer
            Dim MatWK As New Matrix(4)
            If Poly.buf(i).Local.bcmr Then
                MatWK = MatWK.UnitMat()
            Else
                MatWK = Camera.Mat()
                MatWK *= CameraMat
                MatWK = MatWK.InverseMat(dt)
            End If
            Dim MatWK3 As Matrix = MatWK.Clone()
            MatWK *= Poly.buf(i).Local.Mat()
            MatWK = MatWK.Homogeneous()




            '逆行列
            Dim MatWK2 As Matrix = MatWK.Clone()
            Dim Inv As New Matrix(4)
            Inv = MatWK2.InverseMat(dt)
            Inv = Inv.TransposedMat()
            Inv = Inv.Homogeneous()

            'ベクトルワーク
            Dim VecWK As Vector
            Dim VecWK1 As Vector
            Dim VecWK2 As Vector
            Dim VecWK3 As Vector
            '法線ベクトルワーク
            Dim NrmWK1 As Vector
            Dim NrmWK2 As Vector
            Dim VWK As Vector = MatWK.Pos()
            Dim bLbl As Boolean = CheckLabel(VWK, w, h, l, t)
            For IntWK = 0 To Poly.buf(i).v.Length - 1
                NrmWK1 = Inv * Poly.buf(i).v(IntWK).n
                NrmWK1 = NrmWK1.Homogeneous()
                NrmWK2 = MatWK * Poly.buf(i).v(IntWK).n
                NrmWK2 = NrmWK2.Homogeneous()
                VecWK1 = MatWK * Poly.buf(i).v(IntWK).v(0)
                VecWK1 = VecWK1.Homogeneous()
                VecWK2 = MatWK * Poly.buf(i).v(IntWK).v(1)
                VecWK2 = VecWK2.Homogeneous()
                VecWK3 = MatWK * Poly.buf(i).v(IntWK).v(2)
                VecWK3 = VecWK3.Homogeneous()

                'If 0.0 <= VecWK1.x(3) And 0.0 <= VecWK2.x(3) And 0.0 <= VecWK3.x(3) Then
                If VecWK2.x(3) < VecWK1.x(3) And VecWK3.x(3) < VecWK1.x(3) Then
                    VecWK = VecWK1.Clone()
                ElseIf VecWK3.x(3) < VecWK2.x(3) And VecWK1.x(3) < VecWK2.x(3) Then
                    VecWK = VecWK2.Clone()
                Else
                    VecWK = VecWK3.Clone()
                End If

                'ポリゴン描画
                Dim dot As Double = VecWK.Dot(NrmWK1)
                If dot <= 0.0 Or Not b Then
                    DrawTriangle(SB, white, Poly.buf(i).v(IntWK).vcol, NrmWK2, w, h, l, t, VecWK1, VecWK2, VecWK3, Poly.buf(i).v(IntWK).mtr, MatWK3)
                    If Not Poly.buf(i).v(IntWK).tcol = Color.Black And b Then
                        FillTriangle(SB, white, Poly.buf(i).v(IntWK).tcol, NrmWK2, w, h, l, t, VecWK1, VecWK2, VecWK3, Poly.buf(i).v(IntWK).mtr, MatWK3)
                    End If
                End If
                'End If
            Next
            If bLbl Then
                pbwk(i) = True
                pvwk(i) = VWK.Clone()
            End If
        Next

        '球体描画
        For i = 0 To Sphr.Count - 1
            Dim InfWK As New Infinity(Range, Base)

            'オブジェクトローカル座標系
            Dim dt As Double
            Dim MatWK As New Matrix(4)
            If Sphr.buf(i).Local.bcmr Then
                MatWK = MatWK.UnitMat()
            Else
                MatWK = Camera.Mat()
                MatWK *= CameraMat
                MatWK = MatWK.InverseMat(dt)
            End If
            Dim MatWK3 As Matrix = MatWK.Clone()
            MatWK *= Sphr.buf(i).Local.Mat()
            MatWK = MatWK.Homogeneous()


            'ベクトルワーク
            Dim VecWK1 As Vector
            Dim VecWK2 As Vector
            Dim VecWK1F As Vector
            Dim VecWK2F As Vector
            Dim VWK As Vector = MatWK.Pos()
            Dim bLbl As Boolean = CheckLabel(VWK, w, h, l, t)
            VecWK1 = Sphr.buf(i).v.v.Clone()
            VecWK1 = MatWK * VecWK1
            VecWK1.x(1) -= Sphr.buf(i).v.r
            VecWK1.x(2) += Sphr.buf(i).v.r
            VecWK1 = VecWK1.Homogeneous()
            VecWK2 = Sphr.buf(i).v.v.Clone()
            VecWK2 = MatWK * VecWK2
            VecWK2.x(1) += Sphr.buf(i).v.r
            VecWK2.x(2) -= Sphr.buf(i).v.r
            VecWK2 = VecWK2.Homogeneous()


            '透視射影
            If BInf Then
                VecWK1F = InfWK.FrustumInf(VecWK1)
                VecWK2F = InfWK.FrustumInf(VecWK2)
            Else
                VecWK1F = Frs.Frustum(VecWK1)
                VecWK2F = Frs.Frustum(VecWK2)
                VecWK1F.x(0) = 1.0
                VecWK1F.x(1) = -VecWK1F.x(1) / VecWK1F.x(3)
                VecWK1F.x(2) = -VecWK1F.x(2) / VecWK1F.x(3)
                VecWK1F.x(3) = VecWK1.x(3)
                VecWK2F.x(0) = 1.0
                VecWK2F.x(1) = -VecWK2F.x(1) / VecWK2F.x(3)
                VecWK2F.x(2) = -VecWK2F.x(2) / VecWK2F.x(3)
                VecWK2F.x(3) = VecWK2.x(3)
            End If


            '球体描画
            'If Sphr.buf(i).v.r <= VecWK1.x(3) Then
            If Not Sphr.buf(i).v.tcol = Color.Black And b Then
                FillEllipse(SB, white, Sphr.buf(i).v.vcol, w, h, l, t, VecWK1, VecWK2, Sphr.buf(i).v.r, Sphr.buf(i).v.mtr, MatWK3)
            End If
            DrawEllipse(SB, white, Sphr.buf(i).v.vcol, w, h, l, t, VecWK1, VecWK2, Sphr.buf(i).v.r, Sphr.buf(i).v.mtr, MatWK3)
            'End If
            If bLbl Then
                sbwk(i) = True
                svwk(i) = VWK.Clone()
            End If
        Next

        '線分描画
        For i = 0 To Line.Count - 1
            'オブジェクトローカル座標系
            Dim dt As Double
            Dim IntWK As Integer
            Dim MatWK As New Matrix(4)
            If Line.buf(i).Local.bcmr Then
                MatWK = MatWK.UnitMat()
            Else
                MatWK = Camera.Mat()
                MatWK *= CameraMat
                MatWK = MatWK.InverseMat(dt)
            End If
            Dim MatWK3 As Matrix = MatWK.Clone()
            MatWK *= Line.buf(i).Local.Mat()
            MatWK = MatWK.Homogeneous()

            '逆行列
            Dim MatWK2 As Matrix = MatWK.Clone()
            Dim Inv As New Matrix(4)
            Inv = MatWK2.InverseMat(dt)
            Inv = Inv.TransposedMat()
            Inv = Inv.Homogeneous()

            'ベクトルワーク
            Dim NrmWK1 As New Vector(4, True)
            NrmWK1.x(0) = 0.0
            Dim VecWK1 As Vector
            Dim VecWK2 As Vector
            Dim VWK As Vector = MatWK.Pos()
            Dim bLbl As Boolean = CheckLabel(VWK, w, h, l, t)
            '線分描画
            For IntWK = 0 To Line.buf(i).v.Length - 1
                VecWK1 = MatWK * Line.buf(i).v(IntWK).v(0)
                VecWK1 = VecWK1.Homogeneous()
                VecWK2 = MatWK * Line.buf(i).v(IntWK).v(1)
                VecWK2 = VecWK2.Homogeneous()
                If Line.buf(i).Opt Then
                    If VecWK2.x(3) < VecWK1.x(3) Then
                        Dim vcol As Color = Line.buf(i).v(IntWK).vcol
                        vcol.A = vcol.A
                        vcol.R = vcol.R / 3.0
                        vcol.G = vcol.G / 3.0
                        vcol.B = vcol.B / 2.0
                        DrawClippingLine(SB, white, vcol, NrmWK1, w, h, l, t, VecWK1, VecWK2, Line.buf(i).v(IntWK).mtr, MatWK3)
                    Else
                        DrawClippingLine(SB, white, Line.buf(i).v(IntWK).vcol, NrmWK1, w, h, l, t, VecWK1, VecWK2, Line.buf(i).v(IntWK).mtr, MatWK3)
                    End If
                Else
                    If i = 2 And IntWK = 121 Then
                        IntWK = IntWK
                    End If

                    DrawClippingLine(SB, white, Line.buf(i).v(IntWK).vcol, NrmWK1, w, h, l, t, VecWK1, VecWK2, Line.buf(i).v(IntWK).mtr, MatWK3)
                End If
            Next
            If bLbl Then
                lbwk(i) = True
                lvwk(i) = VWK.Clone()
            End If
        Next

        'ラベル描画
        For i = 0 To Poly.Count - 1
            If pbwk(i) Then
                DrawLabel(SB, white, pvwk(i), w, h, l, t, Poly.buf(i).Label)
            End If
        Next
        For i = 0 To Sphr.Count - 1
            If sbwk(i) Then
                DrawLabel(SB, white, svwk(i), w, h, l, t, Sphr.buf(i).Label)
            End If
        Next
        For i = 0 To Line.Count - 1
            If lbwk(i) Then
                DrawLabel(SB, white, lvwk(i), w, h, l, t, Line.buf(i).Label)
            End If
        Next

    End Sub

End Class